from collections import Iterable
import csv

from django.contrib.auth.mixins import PermissionRequiredMixin
from django.core.exceptions import ObjectDoesNotExist
from django.http import HttpResponse
from django.views.generic import ListView

from wafer.talks.models import Talk

from bursary.models import Bursary
from front_desk.models import CheckIn
from register.models import Accomm, Attendee, ChildCare, Meal


class CSVExportView(ListView):
    """Export the given columns for the model as CSV."""
    columns = None
    filename = None

    def get_data_line(self, instance):
        ret = []
        for column in self.columns:
            obj = instance
            for component in column.split('.'):
                try:
                    obj = getattr(obj, component)
                except ObjectDoesNotExist:
                    obj = '%s missing!' % component
                    break
                except AttributeError:
                    obj = getattr(self, component)(obj)
                if not obj:
                    break
                if callable(obj):
                    obj = obj()
            if (not isinstance(obj, (str, bytes))
                    and isinstance(obj, Iterable)):
                ret.extend(str(i) for i in obj)
            else:
                ret.append(str(obj))

        return ret

    def write_rows(self, writer, objects):
        for instance in objects:
            writer.writerow(self.get_data_line(instance))

    def render_to_response(self, context, **response_kwargs):
        response = HttpResponse(content_type='text/csv')
        response['Content-Disposition'] = (
            'attachment; filename="%s"' % self.filename
        )

        writer = csv.writer(response)
        writer.writerow(self.columns)
        self.write_rows(writer, context['object_list'])

        return response


class AttendeeAdminMixin(PermissionRequiredMixin):
    permission_required = 'register.change_attendee'


class AttendeeBadgeExport(AttendeeAdminMixin, CSVExportView):
    model = Attendee
    filename = "attendee_badges.csv"
    ordering = ('user__username',)
    columns = [
        'user.username', 'reconfirm', 'user.email', 'user.get_full_name',
        'nametag_2', 'nametag_3', 'languages', 'food.diet',
    ]


class AttendeeAccommExport(AttendeeAdminMixin, CSVExportView):
    model = Accomm
    filename = "attendee_accommodation.csv"
    ordering = ('attendee__user__username',)
    columns = [
        'attendee.user.username', 'attendee.user.get_full_name',
        'attendee.user.email', 'attendee.reconfirm', 'attendee.paid',
        'attendee.user.bursary.accommodation_status', 'attendee.gender',
        'attendee.country', 'requirements',
        'family_usernames', 'get_checkin_checkouts', 'room',
    ]


class ChildCareExport(AttendeeAdminMixin, CSVExportView):
    model = ChildCare
    filename = "attendee_child_care.csv"
    ordering = ('attendee__user__username',)
    columns = [
        'attendee.user.username', 'attendee.user.get_full_name',
        'attendee.user.email', 'attendee.reconfirm', 'attendee.paid',
        'attendee.user.bursary.accommodation_status',
        'attendee.arrival', 'attendee.departure',
        'needs', 'details',
    ]


class TalksExport(PermissionRequiredMixin, CSVExportView):
    model = Talk
    permission_required = 'talks.edit_private_notes'
    filename = "talk_evaluations.csv"
    ordering = ("talk_id",)
    columns = [
        'talk_id', 'title', 'get_authors_display_name', 'abstract',
        'talk_type.name', 'track.name', 'get_status_display', 'review_score',
        'review_count', 'notes', 'private_notes', 'all_review_comments',
    ]

    def all_review_comments(self, talk):
        return [
            "(%s) %s" % (review.reviewer.username, review.notes.raw)
            for review in talk.reviews.all()
            if review.notes.raw
        ]


class FoodExport(AttendeeAdminMixin, CSVExportView):
    model = Meal
    filename = "meals.csv"
    ordering = ('date', 'meal',)
    columns = [
        'date', 'meal', 'total', 'total_unconfirmed',
        'regular', 'vegetarian', 'vegan', 'gluten_free',
        'other', 'other_details',
    ]

    def render_to_response(self, context, **response_kwargs):
        self._confirmed_attendees = {}
        return super().render_to_response(context, **response_kwargs)

    def attendee_confirmed(self, attendee_id):
        if attendee_id not in self._confirmed_attendees:
            if CheckIn.objects.filter(attendee_id=attendee_id).exists():
                return True

            attendee = Attendee.objects.get(id=attendee_id)
            paid = attendee.paid()

            try:
                bursary = Bursary.objects.get(user=attendee.user)
            except Bursary.DoesNotExist:
                bursary = Bursary()

            confirmed = any((
                not bursary.request_any and attendee.billable() and paid,
                not bursary.request_any and not attendee.billable() and attendee.final_dates,
                bursary.request_any and bursary.status_in(None, ['accepted']),
                attendee.reconfirm,
            ))
            self._confirmed_attendees[attendee_id] = confirmed
        return self._confirmed_attendees[attendee_id]

    def get_data_line(self, meal):
        row = {
            'date': meal.date.isoformat(),
            'meal': meal.meal,
            'total': 0,
            'total_unconfirmed': meal.food_set.count(),
            'regular': 0,
            'vegetarian': 0,
            'vegan': 0,
            'gluten_free': 0,
            'other': 0,
            'other_details': [],
        }

        for food in meal.food_set.all():
            if not self.attendee_confirmed(food.attendee_id):
                continue

            diet = food.diet
            if diet == '':
                diet = 'regular'
            elif diet == 'other':
                details = [food.attendee.user.username]
                details.append(food.special_diet)
                row['other_details'].append(': '.join(details))
            row[diet] += 1
            row['total'] += 1

        row['other_details'] = ', '.join(row['other_details'])
        return [row.get(key) for key in self.columns]


class SpecialDietExport(AttendeeAdminMixin, CSVExportView):
    filename = "diets.csv"
    ordering = ('attendee__user__username',)
    columns = [
        'username', 'name', 'confirmed', 'diet', 'special_diet'
    ]

    def attendee_confirmed(self, attendee):
        try:
            if attendee.check_in:
                return True
        except CheckIn.DoesNotExist:
            pass

        if attendee.reconfirm:
            return True

        paid = attendee.paid()

        try:
            bursary = attendee.user.bursary
        except Bursary.DoesNotExist:
            bursary = Bursary()

        return any((
            not bursary.request_any and attendee.billable() and paid,
            not bursary.request_any and not attendee.billable() and attendee.final_dates,
            bursary.request_any and bursary.status_in(None, ['accepted']),
        ))

    def get_queryset(self):
        return Meal.objects.get(**self.kwargs).food_set.exclude(
                special_diet__exact='',
                diet__exact='',
            ).select_related(
                'attendee',
                'attendee__user',
                'attendee__user__bursary',
                'attendee__user__userprofile',
            ).order_by(*self.ordering)

    def get_data_line(self, food):
        row = {
            'username': food.attendee.user.username,
            'name': food.attendee.user.userprofile.display_name(),
            'confirmed': self.attendee_confirmed(food.attendee),
            'diet': food.diet,
            'special_diet': food.special_diet,
        }
        return [row.get(key) for key in self.columns]

    def write_rows(self, writer, objects):
        super().write_rows(writer, objects)
        count = {True: 0, False: 0}
        for food in Meal.objects.get(**self.kwargs).food_set.filter(
                special_diet__exact='',
                diet__exact='',
            ).select_related(
                'attendee',
                'attendee__user',
                'attendee__user__userprofile',
            ):
            count[self.attendee_confirmed(food.attendee)] += 1
        for confirmed in (True, False):
            row = {
                'username': '*',
                'name': 'Everyone Else - {} people'.format(count[confirmed]),
                'confirmed': confirmed,
                'diet': '',
                'special_diet': '',
            }
            writer.writerow([row.get(key) for key in self.columns])



# -*- coding: utf-8 -*-
# Generated by Django 1.11.18 on 2019-02-04 19:33
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Accomm',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('requirements', models.TextField(blank=True)),
                ('childcare', models.BooleanField()),
                ('childcare_needs', models.TextField(blank=True)),
                ('childcare_details', models.TextField(blank=True)),
                ('special_needs', models.TextField(blank=True)),
                ('family_usernames', models.TextField(blank=True)),
                ('room', models.CharField(blank=True, default='', max_length=128)),
            ],
        ),
        migrations.CreateModel(
            name='AccommNight',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date', models.DateField(unique=True)),
            ],
            options={
                'ordering': ['date'],
            },
        ),
        migrations.CreateModel(
            name='Attendee',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('nametag_2', models.CharField(blank=True, max_length=50)),
                ('nametag_3', models.CharField(blank=True, max_length=50)),
                ('emergency_contact', models.TextField(blank=True)),
                ('announce_me', models.BooleanField()),
                ('register_announce', models.BooleanField()),
                ('register_discuss', models.BooleanField()),
                ('coc_ack', models.BooleanField(default=False)),
                ('fee', models.CharField(blank=True, max_length=5)),
                ('arrival', models.DateTimeField(blank=True, null=True)),
                ('departure', models.DateTimeField(blank=True, null=True)),
                ('final_dates', models.BooleanField(default=False)),
                ('reconfirm', models.BooleanField(default=False)),
                ('t_shirt_size', models.CharField(blank=True, max_length=8)),
                ('shoe_size', models.CharField(blank=True, max_length=8)),
                ('gender', models.CharField(blank=True, max_length=1)),
                ('country', models.CharField(blank=True, max_length=2)),
                ('languages', models.CharField(blank=True, max_length=50)),
                ('pgp_fingerprints', models.TextField(blank=True)),
                ('invoiced_entity', models.TextField(blank=True)),
                ('billing_address', models.TextField(blank=True)),
                ('notes', models.TextField(blank=True)),
                ('completed_register_steps', models.IntegerField(default=0)),
                ('user', models.OneToOneField(on_delete=django.db.models.deletion.PROTECT, related_name='attendee', to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.CreateModel(
            name='Food',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('diet', models.CharField(blank=True, max_length=16)),
                ('special_diet', models.TextField(blank=True)),
                ('attendee', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='food', to='register.Attendee')),
            ],
        ),
        migrations.CreateModel(
            name='Meal',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date', models.DateField(db_index=True)),
                ('meal', models.CharField(max_length=16)),
            ],
            options={
                'ordering': ['date'],
            },
        ),
        migrations.CreateModel(
            name='Queue',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(db_index=True, max_length=32)),
                ('size', models.IntegerField(null=True)),
            ],
            options={
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='QueueSlot',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('timestamp', models.DateTimeField(auto_now_add=True)),
                ('attendee', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='queues', to='register.Attendee')),
                ('queue', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='slots', to='register.Queue')),
            ],
            options={
                'ordering': ['timestamp'],
            },
        ),
        migrations.AlterUniqueTogether(
            name='meal',
            unique_together=set([('date', 'meal')]),
        ),
        migrations.AddField(
            model_name='food',
            name='meals',
            field=models.ManyToManyField(to='register.Meal'),
        ),
        migrations.AddField(
            model_name='accomm',
            name='attendee',
            field=models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='accomm', to='register.Attendee'),
        ),
        migrations.AddField(
            model_name='accomm',
            name='nights',
            field=models.ManyToManyField(to='register.AccommNight'),
        ),
        migrations.AlterUniqueTogether(
            name='queueslot',
            unique_together=set([('attendee', 'queue')]),
        ),
    ]



from collections import Counter, OrderedDict, defaultdict

from django.conf import settings
from django.contrib.auth.mixins import LoginRequiredMixin
from django.core.cache import cache
from django.core.exceptions import ObjectDoesNotExist
from django.views.generic import TemplateView

from bursary.models import Bursary, BURSARY_STATUS_CHOICES
from front_desk.models import CheckIn
from register.models import Attendee, Meal, user_is_registered


def clean_almostdicts(value):
    if not isinstance(value, dict):
        return value

    return OrderedDict(
        (k, clean_almostdicts(v)) for k, v in value.items()
    )


class StatisticsView(LoginRequiredMixin, TemplateView):
    template_name = 'register/statistics.html'
    cache_key = 'register:statistics'
    cache_timeout = 30*60 if not settings.DEBUG else 10

    def get_context_data(self, **kwargs):
        retval = cache.get(self.cache_key)
        if retval:
            return retval

        attendees = Attendee.objects.all()
        attendees_count = len(attendees)
        attendees_registered = 0
        attendees_confirmed = 0
        attendees_arrived = 0
        attendees_by_country = defaultdict(Counter)
        attendees_by_language = defaultdict(Counter)
        attendees_by_gender = defaultdict(Counter)
        fees = defaultdict(Counter)
        tshirts = defaultdict(Counter)
        shoes = defaultdict(Counter)
        accomm_total = 0
        accomm_confirmed = 0
        accommodation = defaultdict(Counter)
        food_total = 0
        food_confirmed = 0
        food_restrictions = defaultdict(Counter)
        food_restrictions['No restrictions']  # get that on top of the list
        meals = defaultdict(
            lambda: defaultdict(Counter)
        )
        for attendee in attendees:
            if attendee.completed_register_steps >= 10:
                attendees_registered += 1
            else:
                continue

            paid = attendee.paid()

            try:
                bursary = Bursary.objects.get(user=attendee.user)
            except Bursary.DoesNotExist:
                bursary = Bursary()

            checked_in = CheckIn.objects.filter(attendee=attendee).exists()

            if checked_in:
                attendees_arrived += 1

            reconfirm = any((
                checked_in,
                not bursary.request_any and attendee.billable() and paid,
                not bursary.request_any and not attendee.billable()
                    and attendee.final_dates,
                bursary.request_any and bursary.status_in(None, ['accepted']),
                attendee.reconfirm,
            ))

            if reconfirm:
                attendees_confirmed += 1

            fees[attendee.fee]['all'] += 1
            if paid:
                fees[attendee.fee]['paid'] += 1

            if attendee.t_shirt_size:
                size = attendee.t_shirt_size
                tshirts[size]['all'] += 1
                if reconfirm:
                    tshirts[size]['confirmed'] += 1

            if attendee.shoe_size:
                size = attendee.shoe_size
                shoes[size]['all'] += 1
                if reconfirm:
                    shoes[size]['confirmed'] += 1

            attendees_by_country[attendee.country]['all'] += 1
            if reconfirm:
                attendees_by_country[attendee.country]['confirmed'] += 1

            attendees_by_gender[attendee.gender]['all'] += 1
            if reconfirm:
                attendees_by_gender[attendee.gender]['confirmed'] += 1

            languages = set(
                attendee.languages.lower()
                .replace(',', ' ')
                .replace('/', ' ')
                .replace(';', ' ')
                .split()
            )
            for language in languages:
                attendees_by_language[language]['all'] += 1
                if reconfirm:
                    attendees_by_language[language]['confirmed'] += 1

            try:
                accomm = attendee.accomm
                if not accomm.nights.exists():
                    raise ObjectDoesNotExist
            except ObjectDoesNotExist:
                pass
            else:
                accomm_total += 1
                if reconfirm:
                    accomm_confirmed += 1

                for night in attendee.accomm.nights.all():
                    accommodation[night]['all'] += 1
                    if reconfirm:
                        accommodation[night]['confirmed'] += 1

            try:
                food = attendee.food
                if not food.meals.exists():
                    raise ObjectDoesNotExist
            except ObjectDoesNotExist:
                pass
            else:
                food_total += 1
                if reconfirm:
                    food_confirmed += 1

                markers = []
                if food.diet:
                    markers.append(food.diet.title())
                if not markers:
                    markers = ['No restrictions']

                food_restrictions[' '.join(markers)]['all'] += 1
                if reconfirm:
                    food_restrictions[' '.join(markers)]['confirmed'] += 1

                for meal in food.meals.all():
                    meals[meal.date][meal.meal]['all'] += 1
                    if reconfirm:
                        meals[meal.date][meal.meal]['confirmed'] += 1

        bursaries_by_status = defaultdict(Counter)
        bursaries_travel = Counter()
        for bursary in Bursary.objects.all():
            if not user_is_registered(bursary.user):
                continue
            for type in ('food', 'accommodation', 'travel'):
                if getattr(bursary, 'request_%s' % type):
                    status = getattr(bursary, '%s_status' % type)
                    bursaries_by_status[type]['all'] += 1
                    bursaries_by_status[type][status] += 1
                    if type == 'travel':
                        amount = bursary.travel_bursary
                        bursaries_travel['all'] += amount
                        bursaries_travel[status] += amount

        # Prepare for presentation
        fees = OrderedDict(
            (label, fees[key])
            for key, label in Attendee.FEES.items()
        )
        attendees_by_country = sorted(
            attendees_by_country.items(), key=lambda x: (-x[1]['all'], x[0])
        )
        attendees_by_gender = sorted(
            attendees_by_gender.items(), key=lambda x: (-x[1]['all'], x[0])
        )
        attendees_by_language = sorted(
            attendees_by_language.items(), key=lambda x: (-x[1]['all'], x[0])
        )
        tshirts = OrderedDict(
            (label, tshirts[key])
            for key, label in settings.DEBCONF_T_SHIRT_SIZES
            if key
        )
        shoes = OrderedDict(
            (size, shoes[size])
            for size, label in settings.DEBCONF_SHOE_SIZES
            if size
        )
        if not settings.DEBCONF_SHOE_SIZES:
            shoes = None
        accommodation = OrderedDict(
            sorted(
                (night.date, counts)
                for night, counts in accommodation.items()
            )
        )
        meal_labels = list(Meal.MEALS.values())
        meals = OrderedDict(
            (day, [day_meals[key] for key in Meal.MEALS])
            for day, day_meals in sorted(meals.items())
        )

        bursary_statuses = ['All'] + [
            choice[0].title() for choice in BURSARY_STATUS_CHOICES
        ]

        bursaries_by_status = OrderedDict(
            (type.title(), OrderedDict(
                (status.lower(), counter[status.lower()])
                for status in bursary_statuses
            ))
            for type, counter in bursaries_by_status.items()
        )
        bursaries_travel = OrderedDict(
            (status, bursaries_travel[status.lower()])
            for status in bursary_statuses
        )

        retval = clean_almostdicts({
            'attendees_count': attendees_count,
            'attendees_registered': attendees_registered,
            'attendees_confirmed': attendees_confirmed,
            'attendees_arrived': attendees_arrived,
            'fees': fees,
            'tshirts': tshirts,
            'shoes': shoes,
            'attendees_by_country': attendees_by_country,
            'attendees_by_language': attendees_by_language,
            'attendees_by_gender': attendees_by_gender,
            'accomm_total': accomm_total,
            'accomm_confirmed': accomm_confirmed,
            'accommodation': accommodation,
            'food_total': food_total,
            'food_confirmed': food_confirmed,
            'food_restrictions': food_restrictions,
            'genders': Attendee.GENDERS,
            'meal_labels': meal_labels,
            'meals': meals,
            'bursary_statuses': bursary_statuses,
            'bursaries_by_status': bursaries_by_status,
            'bursaries_travel': bursaries_travel,
        })

        cache.set(self.cache_key, retval, self.cache_timeout)
        return retval



from django.db import models
from django.utils.functional import cached_property

from register.models.attendee import Attendee


class Queue(models.Model):
    name = models.CharField(max_length=32, db_index=True)
    size = models.IntegerField(null=True)

    def __str__(self):
        return self.name

    class Meta:
        ordering = ['name']


class QueueSlot(models.Model):
    attendee = models.ForeignKey(Attendee, related_name='queues',
                                 on_delete=models.CASCADE)
    queue = models.ForeignKey(Queue, related_name='slots',
                              on_delete=models.CASCADE)
    timestamp = models.DateTimeField(auto_now_add=True)

    @cached_property
    def position(self):
        return QueueSlot.objects.filter(
            queue=self.queue, timestamp__lt=self.timestamp).count() + 1

    def __str__(self):
        return '{}: {} ({})'.format(
            self.queue.name, self.position, self.attendee.user.username)

    class Meta:
        ordering = ['timestamp']
        unique_together = ('attendee', 'queue')



from django.db import models

from register.models import Attendee


class CheckIn(models.Model):
    attendee = models.OneToOneField(Attendee, related_name='check_in',
                                    on_delete=models.PROTECT)
    t_shirt = models.BooleanField(default=False)
    shoes = models.BooleanField(default=False)
    swag = models.BooleanField(default=False)
    nametag = models.BooleanField(default=False)
    key_card = models.CharField(max_length=5, blank=True)
    room_key = models.BooleanField(default=False)
    checked_out = models.BooleanField(default=False)
    returned_key = models.BooleanField(default=False)
    returned_card = models.BooleanField(default=False)
    notes = models.TextField(blank=True)



from django.core.management.base import BaseCommand, CommandError
from django.core.exceptions import ValidationError

from volunteers.models import Task


class Command(BaseCommand):
    help = 'Creates a task'

    def add_arguments(self, parser):
        parser.add_argument('name')
        parser.add_argument('description')
        parser.add_argument('date')
        parser.add_argument('start_time')
        parser.add_argument('end_time')
        parser.add_argument('nbr_volunteers_min', type=int,
                            nargs='?', default=1)
        parser.add_argument('nbr_volunteers_max', type=int,
                            nargs='?', default=1)

    def handle(self, *args, **options):
        try:
            task = Task.objects.create(
                    name=options['name'],
                    description=options['description'],
                    date=options['date'],
                    start_time=options['start_time'],
                    end_time=options['end_time'],
                    nbr_volunteers_min=options['nbr_volunteers_min'],
                    nbr_volunteers_max=options['nbr_volunteers_max'])
        except ValidationError as e:
            raise CommandError('Invalid task: "%s"' % e)

        task.save()

        self.stdout.write('Successfully created task: "%s"' % task)



from django import forms
from django.core.signing import BadSignature

from crispy_forms.helper import FormHelper

from invoices.models import Invoice


class InvoiceForm(forms.Form):
    url = forms.URLField(
        label='URL',
        required=False,
    )

    def __init__(self, *args, **kwargs):
        self.user = kwargs.pop('user', None)
        super().__init__(*args, **kwargs)

        self.helper = FormHelper()
        self.helper.form_tag = False
        self.helper.include_media = False

    def clean_url(self):
        url = self.cleaned_data['url']
        invalid_url_error = 'Needs to a be a valid URL for a visible invoice'

        # The URL validator should have ensured that we have enough slashes to
        # do things like this
        reference_number = url.rstrip('/').rsplit('/', 1)[1]
        signed_url = ':' in reference_number
        if signed_url:
            try:
                reference_number = Invoice.signer.unsign(reference_number)
            except BadSignature:
                raise forms.ValidationError(invalid_url_error)

        try:
            invoice = Invoice.objects.get(reference_number=reference_number)
        except Invoice.DoesNotExist:
            raise forms.ValidationError(invalid_url_error)

        if self.user != invoice.recipient and not signed_url:
            raise forms.ValidationError(invalid_url_error)

        if invoice.status != 'new':
            raise forms.ValidationError('Invoice is not unpaid')

        if invoice.compound:
            raise forms.ValidationError('This is a combined invoice, itself')

        self.invoice = invoice

        return url


class InvoiceFormset(forms.BaseFormSet):
    def clean(self):
        if any(self.errors):
            return

        invoices = []
        for form in self.forms:
            if hasattr(form, 'invoice'):
                invoices.append(form.invoice.pk)

        if len(invoices) > len(set(invoices)):
            raise forms.ValidationError('Invoices must be distinct')

        if len(invoices) < 2:
            raise forms.ValidationError(
                'At least 2 URLs need to be provided for any combining to '
                'happen.')


class CombineForm(forms.Form):
    invoiced_entity = forms.CharField(
        label='Invoiced entity',
    )
    billing_address = forms.CharField(
        label='Billing address',
        widget=forms.Textarea(attrs={'rows': 3}),
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.helper = FormHelper()
        self.helper.form_tag = False
        self.helper.include_media = False



from django import forms
from django.conf import settings

from crispy_forms.helper import FormHelper
from crispy_forms.layout import Field, Fieldset, Layout, HTML

from invoices.prices import ACCOMM_INVOICE_INFO
from register.dates import night_choices
from register.fields import NightSelectionField
from bursary.models import Bursary


class AccommodationForm(forms.Form):
    accomm = forms.BooleanField(
        label='I need conference-organised accommodation',
        widget=forms.Select(choices=(
            (False, 'No, I will find my own accommodation'),
            (True, 'Yes, I need accommodation'),
        )),
        required=False,
    )
    nights = forms.MultipleChoiceField(
        label="I'm requesting accommodation for these nights:",
        help_text='The "night of" is the date of the day before a night. '
                  'So accommodation on the night of 6 Aug ends on the '
                  'morning of the 7th.',
        choices=night_choices(),
        widget=forms.CheckboxSelectMultiple,
        required=False,
    )
    requirements = forms.CharField(
        label='Do you have any particular accommodation requirements?',
        help_text='Anything that you want us to consider for room attribution '
                  'should be listed here (ex. "I want to be with Joe Hill", '
                  '"I snore", "I go to bed early", "I need wheelchair access")',
        widget=forms.Textarea(attrs={'rows': 3}),
        required=False,
    )
    family_usernames = forms.CharField(
        label='Usernames of my family members, '
              'who have registered separately',
        help_text="One per line. This isn't validated.",
        widget=forms.Textarea(attrs={'rows': 3}),
        required=False,
    )

    def __init__(self, *args, **kwargs):
        attendee = kwargs.pop('attendee')
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.form_tag = False
        self.helper.include_media = False

        try:
            self.bursary = attendee.user.bursary
        except Bursary.DoesNotExist:
            self.bursary = Bursary()

        if settings.DEBCONF_PAID_ACCOMMODATION:
            accomm_availability = HTML(
                '<p>The cost is {} USD/night for attendees who do not '
                'receive a bursary.</p>'.format(
                    ACCOMM_INVOICE_INFO['unit_price']))
        else:
            accomm_availability = HTML(
                '<p>Conference-organised accommodation is only available to '
                'attendees who receive an accommodation bursary. If you are '
                'paying for your own accommodation, see our '
                '<a href="/about/accommodation">list of nearby options</a>.')

        if (settings.DEBCONF_PAID_ACCOMMODATION
                or self.bursary.request_accommodation):
            self.helper.layout = Layout(
                accomm_availability,
                Field('accomm', id='accomm'),
                Fieldset(
                    'Accommodation Details',
                    NightSelectionField('nights'),
                    'requirements',
                    Field('family_usernames'),
                    css_id='accomm-details',
                )
            )
        else:
            self.helper.layout = Layout(accomm_availability)

    def clean(self):
        cleaned_data = super().clean()

        if not cleaned_data.get('accomm'):
            if self.bursary.request_accommodation:
                self.add_error('accomm',
                    'Accommodation bursary was requested, but no accommodation '
                    'selected.')
            return cleaned_data

        if (not settings.DEBCONF_PAID_ACCOMMODATION and
                not self.bursary.request_accommodation):
            self.add_error(
                'accomm',
                'Accommodation is only available to attendees who receive an '
                'accommodation bursary.')

        if not cleaned_data.get('nights'):
            self.add_error(
                'accomm',
                'Please select the nights you require accommodation for.')

        return cleaned_data



from register.views.accommodation import AccommodationView
from register.views.attendee import ContactInformationView
from register.views.billing import BillingView
from register.views.bursary import BursaryView
from register.views.childcare import ChildCareView
from register.views.conference import ConferenceRegistrationView
from register.views.confirmation import ConfirmationView
from register.views.food import FoodView
from register.views.instructions import InstructionsView
from register.views.misc import MiscView
from register.views.personal import PersonalInformationView
from register.views.review import ReviewView


STEPS = [
    InstructionsView,
    ContactInformationView,
    ConferenceRegistrationView,
    PersonalInformationView,
    BursaryView,
    FoodView,
    AccommodationView,
    ChildCareView,
    MiscView,
    ReviewView,
    BillingView,
    ConfirmationView,
]



from django.conf.urls import url

from exports.views import (
    AttendeeAccommExport, AttendeeBadgeExport, ChildCareExport, FoodExport,
    TalksExport, SpecialDietExport,
)

urlpatterns = [
    url(r'^attendees/admin/export/accomm/$', AttendeeAccommExport.as_view(),
        name='attendee_admin_export_accomm'),
    url(r'^attendees/admin/export/badges/$', AttendeeBadgeExport.as_view(),
        name='attendee_admin_export_badges'),
    url(r'^attendees/admin/export/food/$', FoodExport.as_view()),
    url(r'^attendees/admin/export/special_diets/'
        r'(?P<date>[0-9-]+)/(?P<meal>[a-z]+)/$', SpecialDietExport.as_view()),
    url(r'^attendees/admin/export/child_care/$', ChildCareExport.as_view()),
    url(r'^talks/admin/export/$', TalksExport.as_view(),
        name='talks_admin_export'),
]



from django.conf import settings
from django.core.management.base import BaseCommand

from wafer.talks.models import Track, TalkType

import yaml


class Command(BaseCommand):
    help = 'Load Track and Talk Type definitions from YAML files into the DB'

    def handle(self, *args, **options):
        with open(settings.TRACKS_FILE) as f:
            tracks = yaml.safe_load(f)
        with open(settings.TALK_TYPES_FILE) as f:
            talk_types = yaml.safe_load(f)

        for item in tracks:
            self.__load__(Track, item)

        for item in talk_types:
            self.__load__(TalkType, item)

    def __load__(self, model, data):
        name = data.pop('name')
        model.objects.update_or_create(name=name, defaults=data)
        self.stdout.write('Loaded %s %s\n' % (model.__name__, name))




import re

from django.conf import settings
from django import forms
from django.core.exceptions import ValidationError

from crispy_forms.helper import FormHelper
from crispy_forms.layout import Field, Layout
from django_countries import Countries
from django_countries.fields import LazyTypedChoiceField

from register.models.attendee import Attendee


T_SHIRT_CHART_LINK = '<a href="/about/t-shirts">t-shirt sizes chart</a>'
SHOE_CHART_LINK = '<a href="/about/shoe-sizes">shoe size chart</a>'


class OptionalCountries(Countries):
    first = ('__',)
    override = {
        '__': 'Decline to state',
        'XK': 'Kosovo',
    }


class PersonalInformationForm(forms.Form):
    t_shirt_size = forms.ChoiceField(
        label='My t-shirt size',
        choices=settings.DEBCONF_T_SHIRT_SIZES,
        help_text='Refer to the ' + T_SHIRT_CHART_LINK + '.',
        required=False,
    )
    shoe_size = forms.ChoiceField(
        label='My shoe size',
        choices=settings.DEBCONF_SHOE_SIZES,
        help_text='Refer to the ' + SHOE_CHART_LINK + '.',
        required=False,
    )
    gender = forms.ChoiceField(
        label='My gender',
        choices=Attendee.GENDERS.items(),
        help_text='For diversity statistics.',
        required=False,
    )
    country = LazyTypedChoiceField(
        label='The country I call home',
        help_text='For diversity statistics.',
        choices=OptionalCountries(),
        required=False,
    )
    languages = forms.CharField(
        label='The languages I speak',
        help_text='We will list these on your name-tag.',
        initial='en',
        max_length=50,
        required=False,
    )
    pgp_fingerprints = forms.CharField(
        label='My PGP key fingerprints for keysigning',
        help_text='Optional. One fingerprint per line, if you want to take '
                  'part in <a href="https://wiki.debconf.org/wiki/DebConf18/'
                  'Keysigning">the continuous keysigning party</a>. '
                  'Will appear on nametags and a keysigning notesheet, with '
                  'no verification done by the conference organisers.<br>'
                  'Just provide the fingerprints, e.g. <code>1234 5678 90AB '
                  'CDEF 0000  1111 2222 3333 4444 5555</code>',
        widget=forms.Textarea(attrs={'rows': 2}),
        required=False,
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.form_tag = False
        self.helper.include_media = False
        if settings.DEBCONF_SHOE_SIZES:
            shoe_size = (Field('shoe_size'),)
        else:
            shoe_size = ()
        self.helper.layout = Layout(
            Field('t_shirt_size'),
            *shoe_size,
            Field('gender'),
            Field('country'),
            Field('languages'),
            Field('pgp_fingerprints'),
        )

    def clean_pgp_fingerprints(self):
        fingerprints = self.cleaned_data.get('pgp_fingerprints').strip()
        if not fingerprints:
            return ''
        if '-----BEGIN' in fingerprints:
            raise ValidationError('Only fingerprints, not keys, please')
        cleaned = []
        for line in fingerprints.splitlines():
            fp = line.strip().upper().replace(' ', '')
            if fp.startswith('0x'):
                fp = fp[2:]
            if not re.match(r'^[0-9A-F]{40}$', fp):
                raise ValidationError(
                    '{} is not a PGP fingerprint'.format(line))
            cleaned.append('{} {} {} {} {}  {} {} {} {} {}'.format(
                *[fp[i:i + 4] for i in range(0, 40, 4)]))
        return '\n'.join(cleaned)



from django import template
from django.conf import settings

register = template.Library()

debconf = {
    'LOCAL_CURRENCY': settings.DEBCONF_LOCAL_CURRENCY,
}

@register.simple_tag
def debconf_setting(key):
    return debconf[key]



from register.dates import conference_dinner_meal, parse_date
from register.forms.food import FoodForm
from register.models.food import Food, Meal
from register.models.queue import Queue
from register.views.core import RegisterStep


class FoodView(RegisterStep):
    title = 'Food'
    template_name = 'register/page/food.html'
    form_class = FoodForm

    def get_initial(self):
        user = self.request.user
        initial = {}

        try:
            food = user.attendee.food
        except Food.DoesNotExist:
            return initial

        for field in food._meta.get_fields():
            if field.is_relation:
                continue
            initial[field.name] = getattr(food, field.name)

        initial['meals'] = [meal.form_name for meal in food.meals.all()
                            if not meal.conference_dinner]
        initial['conference_dinner'] = any(meal for meal in food.meals.all()
                                           if meal.conference_dinner)
        return initial

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs['attendee'] = self.request.user.attendee
        return kwargs

    def form_valid(self, form):
        attendee = self.request.user.attendee
        data = form.cleaned_data.copy()

        meals = data.pop('meals')

        if not meals:
            Food.objects.filter(attendee=attendee).delete()
            return super().form_valid(form)

        food, created = Food.objects.update_or_create(
            attendee=attendee, defaults=data)
        attendee.food = food

        stored_meals = set(food.meals.all())
        requested_meals = set()
        for meal in meals:
            meal, date = meal.split('_')
            date = parse_date(date)
            requested_meals.add(Meal.objects.get(meal=meal, date=date))

        food.meals.remove(*(stored_meals - requested_meals))
        food.meals.add(*(requested_meals - stored_meals))

        conference_dinner = conference_dinner_meal() in meals
        if conference_dinner:
            queue, created = Queue.objects.get_or_create(
                name='Conference Dinner')
            queue.slots.get_or_create(attendee=attendee)

        return super().form_valid(form)



import datetime

from django.contrib import admin
from django.contrib.admin.templatetags.admin_list import _boolean_icon

from register.admin.core import NotNullFilter
from register.dates import get_ranges_for_dates
from register.models.accommodation import Accomm


class RequirementsNotNullFilter(NotNullFilter):
    title = "Special Requirements"
    parameter_name = "requirements"


class AccommAdmin(admin.ModelAdmin):
    list_display = (
        'attendee', 'full_name', 'email',
        'dates', 'reconfirm', 'bursary', 'room'
    )
    list_editable = ('room',)
    list_filter = (
        'nights', RequirementsNotNullFilter,
        'attendee__reconfirm', 'attendee__user__bursary__accommodation_status',
    )
    search_fields = (
        'attendee__user__username', 'attendee__user__first_name',
        'attendee__user__last_name'
    )

    def full_name(self, instance):
        return instance.attendee.user.get_full_name()
    full_name.admin_order_field = 'attendee__user__last_name'

    def email(self, instance):
        return instance.attendee.user.email
    email.admin_order_field = 'attendee__user__email'

    def reconfirm(self, instance):
        return _boolean_icon(instance.attendee.reconfirm)
    reconfirm.short_description = 'Confirmed?'
    reconfirm.admin_order_field = 'attendee__reconfirm'

    def bursary(self, instance):
        return instance.attendee.user.bursary.accommodation_status
    bursary.short_description = 'Accomm bursary status'
    bursary.admin_order_field = 'attendee__user__bursary__accommodation_status'

    def dates(self, instance):
        to_show = []
        stays = get_ranges_for_dates(
            night.date for night in instance.nights.all()
        )
        for first_night, last_night in stays:
            last_morning = last_night + datetime.timedelta(days=1)
            num_nights = (last_morning - first_night).days
            to_show.append("%s eve. to %s morn. (%s nights)" % (
                first_night, last_morning, num_nights
            ))
        return '; '.join(to_show)


admin.site.register(Accomm, AccommAdmin)



import re

from django.contrib.staticfiles.templatetags.staticfiles import static

from markdown.extensions import Extension
from markdown.postprocessors import Postprocessor

assetRE = re.compile(r'{% static [\'"](.*?)[\'"] %}')


class DjangoStaticAssetsProcessor(Postprocessor):
    def replacement(self, match):
        path = match.group(1)
        url = static(path)
        return url

    def run(self, text):
        return assetRE.sub(self.replacement, text)


class DCExtension(Extension):
    def extendMarkdown(self, md, md_globals):
        md.postprocessors.add('django_static_assets',
                              DjangoStaticAssetsProcessor(md), '_end')


def makeExtension(**kwargs):
    return DCExtension(**kwargs)



from collections import OrderedDict, defaultdict
import datetime
from itertools import repeat
import logging

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.mixins import PermissionRequiredMixin
from django.db import transaction
from django.urls import reverse
from django.views.generic.base import TemplateView
from django.views.generic.detail import SingleObjectMixin
from django.views.generic.edit import FormView

from invoices.prices import MEAL_PRICES, invoice_user
from invoices.models import Invoice
from register.dates import (
    ARRIVE_ON_OR_AFTER, LEAVE_ON_OR_BEFORE, T_SHIRT_SWAP_ON_OR_AFTER,
    parse_date, meals as dc_meals)
from register.models import Accomm, Attendee, Food, Meal

from front_desk.forms import (
    CashInvoicePaymentForm, CheckInForm, CheckOutForm, FoodForm,
    RegisterOnSiteForm, SearchForm, ShoesForm, TShirtForm)
from front_desk.models import CheckIn

log = logging.getLogger(__name__)
shirt_sizes = dict(settings.DEBCONF_T_SHIRT_SIZES)


class CheckInPermissionMixin(PermissionRequiredMixin):
    permission_required = 'front_desk.change_checkin'


class CashBoxPermissionMixin(PermissionRequiredMixin):
    permission_required = 'invoices.change_invoice'


class Dashboard(CheckInPermissionMixin, TemplateView):
    template_name = 'front_desk/dashboard.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        form = SearchForm(data=self.request.GET)

        q = None
        if form.is_valid():
            q = form.cleaned_data['q']

        results = self.search_attendees(q)
        # Strip duplicates while maintaining order
        results = list(OrderedDict(zip(results, repeat(None))))

        context.update({
            'view': self,
            'form': form,
            'results': results[:20],
            'num_results': len(results),
        })
        return context

    def search_attendees(self, query):
        if not query:
            return

        yield from Attendee.objects.filter(user__username=query)

        if '@' in query:
            yield from Attendee.objects.filter(user__email=query)

        if ' ' in query:
            first, last = query.split(' ', 1)
            yield from Attendee.objects.filter(
                user__first_name__iexact=first,
                user__last_name__iexact=last,
            )
        yield from Attendee.objects.filter(user__username__icontains=query)
        yield from Attendee.objects.filter(user__email__icontains=query)
        yield from Attendee.objects.filter(user__first_name__icontains=query)
        yield from Attendee.objects.filter(user__last_name__icontains=query)


class CheckInView(CheckInPermissionMixin, SingleObjectMixin, FormView):
    template_name = 'front_desk/check_in.html'
    form_class = CheckInForm

    model = Attendee
    slug_field = 'user__username'
    slug_url_kwarg = 'username'

    def get(self, request, *args, **kwargs):
        self.object = self.get_object()
        return super().get(request, *args, **kwargs)

    @transaction.atomic
    def post(self, request, *args, **kwargs):
        self.object = self.get_object()
        return super().post(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        attendee = self.object
        initial_tshirt = {
            'size': attendee.t_shirt_size,
        }
        initial_shoes = {
            'size': attendee.shoe_size,
        }

        t_shirt = 'No T-Shirt'
        if attendee.t_shirt_size:
            t_shirt = shirt_sizes[attendee.t_shirt_size]

        try:
            accomm = attendee.accomm
        except Accomm.DoesNotExist:
            accomm = None

        dates = sorted(set(date for meal, date in dc_meals(orga=True)))
        meal_labels = ['Breakfast', 'Lunch', 'Dinner']

        try:
            food = attendee.food
        except Food.DoesNotExist:
            meals = None
        else:
            meals = []
            for meal, meal_label in Meal.MEALS.items():
                cur_meal = []
                meal_dates = [entry.date
                              for entry in food.meals.filter(meal=meal)]
                for d in dates:
                    cur_meal.append((d, d in meal_dates))
                meals.append((meal_label, cur_meal))

        today = datetime.date.today()
        context.update({
            'accomm': accomm,
            'dates': dates,
            'DEBCONF_LOCAL_CURRENCY': settings.DEBCONF_LOCAL_CURRENCY,
            'DEBCONF_SHOE_SIZES': settings.DEBCONF_SHOE_SIZES,
            'invoices': Invoice.objects.filter(recipient=attendee.user),
            'meal_labels': meal_labels,
            'meals': meals,
            't_shirt': t_shirt,
            'shoes': attendee.shoe_size,
            't_shirt_form': TShirtForm(
                initial=initial_tshirt,
                username=attendee.user.username),
            'shoes_form': ShoesForm(
                initial=initial_shoes,
                username=attendee.user.username),
            't_shirt_swap_available': T_SHIRT_SWAP_ON_OR_AFTER <= today,
            'T_SHIRT_SWAP_ON_OR_AFTER': T_SHIRT_SWAP_ON_OR_AFTER,
        })
        return context

    def get_initial(self):
        try:
            check_in = self.object.check_in
        except CheckIn.DoesNotExist:
            return {}
        return {
            't_shirt': check_in.t_shirt,
            'shoes': check_in.shoes,
            'nametag': check_in.nametag,
            'notes': check_in.notes,
            'room_key': check_in.room_key,
            'key_card': check_in.key_card,
            'swag': check_in.swag,
        }

    def form_valid(self, form):
        attendee = self.object

        check_in, created = CheckIn.objects.update_or_create(
            attendee=attendee, defaults=form.cleaned_data)
        attendee.check_in = check_in

        log.info('Checked-in %s. Room Key: %s, Card: %s',
                 attendee.user.username, check_in.room_key, check_in.key_card)

        return super().form_valid(form)

    def get_success_url(self):
        return reverse('front_desk')


class CheckOutView(CheckInPermissionMixin, SingleObjectMixin, FormView):
    template_name = 'front_desk/check_out.html'
    form_class = CheckOutForm

    model = Attendee
    slug_field = 'user__username'
    slug_url_kwarg = 'username'

    def get(self, request, *args, **kwargs):
        self.object = self.get_object()
        return super().get(request, *args, **kwargs)

    @transaction.atomic
    def post(self, request, *args, **kwargs):
        self.object = self.get_object()
        return super().post(request, *args, **kwargs)

    def get_initial(self):
        check_in = self.object.check_in
        return {
            'returned_key': check_in.returned_key,
            'returned_card': check_in.returned_card,
            'notes': check_in.notes,
        }

    def form_valid(self, form):
        attendee = self.object

        data = form.cleaned_data
        data['checked_out'] = True

        check_in, created = CheckIn.objects.update_or_create(
            attendee=attendee, defaults=data)
        attendee.check_in = check_in

        log.info('Checked-Out %s. Returned Key: %s, Card: %s',
                 attendee.user.username, check_in.returned_key,
                 check_in.returned_card)

        return super().form_valid(form)

    def get_success_url(self):
        return reverse('front_desk')


class ChangeFoodView(CashBoxPermissionMixin, SingleObjectMixin, FormView):
    form_class = FoodForm
    template_name = 'front_desk/change_food.html'

    model = Attendee
    slug_field = 'user__username'
    slug_url_kwarg = 'username'

    def get(self, request, *args, **kwargs):
        self.object = self.get_object()
        self.username = self.kwargs['username']
        return super().get(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['prices'] = MEAL_PRICES
        return context

    @transaction.atomic
    def post(self, request, *args, **kwargs):
        self.object = self.get_object()
        self.username = self.kwargs['username']
        return super().post(request, *args, **kwargs)

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs['instance'] = self.object
        try:
            food = self.object.food
        except Food.DoesNotExist:
            pass
        else:
            kwargs['initial'] = {
                'meals': [meal.form_name for meal in food.meals.all()],
            }
        return kwargs

    def get_success_url(self):
        return reverse('front_desk.check_in',
                       kwargs={'username': self.username})

    def form_valid(self, form):
        data = form.cleaned_data
        user = self.object.user
        Invoice.objects.filter(
            recipient=user, status='new').update(status='canceled')
        if data['meals']:
            food, created = Food.objects.get_or_create(attendee=self.object)
            stored_meals = set(food.meals.all())
            requested_meals = set()
            for meal in data['meals']:
                meal, date = meal.split('_')
                date = parse_date(date)
                requested_meals.add(Meal.objects.get(meal=meal, date=date))

            added_meals = requested_meals - stored_meals
            removed_meals = stored_meals - requested_meals
            food.meals.remove(*removed_meals)
            food.meals.add(*added_meals)
            log.info('Modified meals for %s: Adding %s. Removing %s',
                     self.username,
                     ', '.join(sorted(str(meal) for meal in added_meals)),
                     ', '.join(sorted(str(meal) for meal in removed_meals)))
            invoice_user(user, save=True)
        else:
            Food.objects.filter(attendee=self.object).delete()
            log.info('Cancelled all meals for %s', self.username)


        return super().form_valid(form)


class ChangeShirtView(CheckInPermissionMixin, FormView):
    form_class = TShirtForm

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        self.username = self.kwargs['username']
        kwargs['username'] = self.username
        return kwargs

    def get_success_url(self):
        return reverse('front_desk.check_in',
                       kwargs={'username': self.username})

    def update_size(self, attendee, size):
        log.info('T-Shirt swap for %s: %s -> %s',
                 self.username, attendee.t_shirt_size, size)

        attendee.t_shirt_size = size
        attendee.save()

    @transaction.atomic
    def form_valid(self, form):
        attendee = Attendee.objects.get(user__username=self.username)
        size = form.cleaned_data['size']
        self.update_size(attendee, size)
        return super().form_valid(form)


class ChangeShoesView(ChangeShirtView):
    form_class = ShoesForm

    def update_size(self, attendee, size):
        log.info('Shoe swap for %s: %s -> %s',
                 self.username, attendee.shoe_size, size)

        attendee.shoe_size = size
        attendee.save()


class RegisterOnSite(CheckInPermissionMixin, FormView):
    form_class = RegisterOnSiteForm
    template_name = 'front_desk/register.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update({
            'ARRIVE_ON_OR_AFTER': ARRIVE_ON_OR_AFTER,
            'LEAVE_ON_OR_BEFORE': LEAVE_ON_OR_BEFORE,
        })
        return context

    @transaction.atomic
    def form_valid(self, form):
        username = form.cleaned_data['username']
        self.username = username
        email = form.cleaned_data['email']

        names = form.cleaned_data['name'].split(None, 1)
        if len(names) > 1:
            first_name, last_name = names
        else:
            first_name, last_name = names[0], ''

        user = get_user_model().objects.create_user(
            username=username, email=email, first_name=first_name,
            last_name=last_name)
        Attendee.objects.create(
            user=user, nametag_3=username, arrival=datetime.datetime.now(),
            departure=form.cleaned_data['departure'], announce_me=False,
            register_announce=False, register_discuss=False,
            final_dates=True, reconfirm=True)

        log.info('Registered on-site: %s', username)

        return super().form_valid(form)

    def get_success_url(self):
        return reverse('front_desk.check_in',
                       kwargs={'username': self.username})


class CashInvoicePayment(CashBoxPermissionMixin, SingleObjectMixin, FormView):
    form_class = CashInvoicePaymentForm
    template_name = 'front_desk/cash_invoice_payment.html'

    model = Invoice
    slug_field = 'reference_number'
    slug_url_kwarg = 'ref'

    def get(self, request, *args, **kwargs):
        self.object = self.get_object()
        return super().get(request, *args, **kwargs)

    @transaction.atomic
    def post(self, request, *args, **kwargs):
        self.object = self.get_object()
        return super().post(request, *args, **kwargs)

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs['instance'] = self.object
        return kwargs

    def form_valid(self, form):
        invoice = self.object
        invoice.transaction_id = 'Cash: {}'.format(
            form.cleaned_data['receipt_number'])
        invoice.status = 'paid'
        invoice.save()

        log.info('Recieved cash-payment for %s: %s',
                 invoice.recipient.username, invoice.transaction_id)

        return super().form_valid(form)

    def get_success_url(self):
        return reverse('front_desk.check_in',
                       kwargs={'username': self.object.recipient.username})


class RoomsView(CheckInPermissionMixin, TemplateView):
    template_name = 'front_desk/rooms.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        rooms = defaultdict(list)
        for accomm in Accomm.objects.select_related(
                'attendee', 'attendee__user',
                'attendee__user__userprofile',
                'attendee__check_in').all():
            room = accomm.room
            if not room:
                room = None
            rooms[room].append(accomm.attendee)

        unassigned = rooms.pop(None, [])

        if rooms:
            num_beds = max(len(occupants) for room, occupants in rooms.items())
        else:
            num_beds = 0
        bed_names = ['Bed {}'.format(i+1) for i in range(num_beds)]

        for room, occupants in rooms.items():
            occupants += [None] * (num_beds - len(occupants))

        context.update({
            'rooms': sorted(rooms.items()),
            'num_beds': num_beds,
            'bed_names': bed_names,
            'unassigned': unassigned,
        })
        return context
